# pylint: disable=missing-module-docstring

import csv
import warnings
from ast import literal_eval

import numpy as np

from hrosailing.pipelinecomponents import (
    ArithmeticMeanInterpolator,
    Ball,
    WeightedPoints,
)
from hrosailing.wind import convert_apparent_wind_to_true

from ._basepolardiagram import PolarDiagram, PolarDiagramException
from ._plotting import (
    plot3d,
    plot_color_gradient,
    plot_convex_hull,
    plot_flat,
    plot_polar,
)


class PolarDiagramPointcloud(PolarDiagram):
    """A class to represent, visualize and work with a polar diagram
    given by a point cloud.

    Parameters
    ----------
    points : array_like of shape (n, 3)
        Initial points of the point cloud, given as a sequence of
        points consisting of wind speed, wind angle and
        boat speed.
        Points with negative wind speeds will be ignored.

    apparent_wind : bool, optional
        Specifies if wind data is given in apparent wind.

        If `True`, data will be converted to true wind.

        Defaults to `False`.
    """

    def __init__(self, points, apparent_wind=False):
        if apparent_wind:
            points = convert_apparent_wind_to_true(points)
        else:
            points = np.asarray_chkfinite(points)
            points = points[np.where(points[:, 0] >= 0)]
            # if np.any((points[:, 0] <= 0)):
            #    raise PolarDiagramInitializationException(
            #        "`points` has non-positive wind speeds"
            #    )
            points[:, 1] %= 360

        self._points = points

    def __str__(self):
        table = ["   TWS      TWA     BSP\n", "++++++  +++++++  ++++++\n"]
        for point in self.points:
            for i in range(3):
                entry = f"{float(point[i]):.2f}"
                if i == 1:
                    table.append(entry.rjust(7))
                    table.append("  ")
                    continue

                table.append(entry.rjust(6))
                table.append("  ")
            table.append("\n")
        return "".join(table)

    def __repr__(self):
        return f"PolarDiagramPointcloud(pts={self.points})"

    def __call__(
        self,
        ws,
        wa,
        interpolator=ArithmeticMeanInterpolator(50),
        neighbourhood=Ball(radius=1),
    ):
        """Returns the value of the polar diagram at a given `ws-wa` point.

        If the `ws-wa` point is in the cloud, the corresponding boat speed is
        returned, otherwise the value is interpolated.

        Parameters
        ----------
        ws : scalar
            Wind speed.

        wa : scalar
            Wind angle.

        interpolator : Interpolator, optional
            Interpolator subclass that determines the interpolation
            method used to determine the value at the `ws-wa` point.

            Defaults to `ArithmeticMeanInterpolator(50)`.

        neighbourhood : Neighbourhood, optional
            Neighbourhood subclass used to determine the points in
            the point cloud that will be used in the interpolation.

            Defaults to `Ball(radius=1)`.

        Returns
        -------
        bsp : scalar
            Boat speed value as determined above.
        """
        if np.any((ws <= 0)):
            raise PolarDiagramException("`ws` is non-positive")

        wa %= 360

        cloud = self.points
        point = cloud[np.logical_and(cloud[:, 0] == ws, cloud[:, 1] == wa)]
        if point.size:
            return point[2]

        point = np.array([ws, wa])
        weighted_points = WeightedPoints(data=cloud, weights=1)

        considered_points = neighbourhood.is_contained_in(cloud[:, :2] - point)

        return interpolator.interpolate(
            weighted_points[considered_points], point
        )

    @property
    def wind_speeds(self):
        """Returns all unique wind speeds in the point cloud."""
        return np.array(sorted(list(set(self.points[:, 0]))))

    @property
    def wind_angles(self):
        """Returns all unique wind angles in the point cloud."""
        return np.array(sorted(list(set(self.points[:, 1]))))

    @property
    def boat_speeds(self):
        """Returns all occurring boat speeds in the point cloud
        (including duplicates).
        """
        return self.points[:, 2]

    @property
    def points(self):
        """Returns a read only version of `self._points`."""
        return self._points.copy()

    def to_csv(self, csv_path):
        """Creates a .csv file with delimiter ',' and the
        following format:

            `PolarDiagramPointcloud`
            TWS,TWA,BSP
            `self.points`

        Parameters
        ----------
        csv_path : path-like
            Path to a .csv-file or where a new .csv file will be created.
        """
        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.writer(file, delimiter=",")
            csv_writer.writerow([self.__class__.__name__])
            csv_writer.writerow(["TWS", "TWA", "BSP"])
            csv_writer.writerows(self.points)

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=",")
        next(csv_reader)
        points = np.array(
            [[literal_eval(point) for point in row] for row in csv_reader]
        )

        return PolarDiagramPointcloud(points)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram,
        by mirroring it at the 0° - 180° axis and returning a new instance.

        Warning
        -------
        Should only be used if all the wind angles of the initial
        polar diagram are on one side of the 0° - 180° axis,
        otherwise this can result in the construction of duplicate points,
        that might overwrite or live alongside old points.
        """
        if not self.points.size:
            return self

        below_180 = [wa for wa in self.wind_angles if wa <= 180]
        above_180 = [wa for wa in self.wind_angles if wa > 180]
        if below_180 and above_180:
            warnings.warn(
                "there are wind angles on both sides of the 0° - 180° axis. "
                "This might result in duplicate data, "
                "which can overwrite or live alongside old data"
            )

        mirrored_points = self.points
        mirrored_points[:, 1] = 360 - mirrored_points[:, 1]
        symmetric_points = np.row_stack((self.points, mirrored_points))

        return PolarDiagramPointcloud(symmetric_points)

    def add_points(self, new_pts, apparent_wind=False):
        """Adds additional points to the point cloud.

        Parameters
        ----------
        new_pts : array_like of shape (n, 3)
            New points to be added to the point cloud given as a sequence
            of points consisting of wind speed, wind angle and
            boat speed.

        apparent_wind : bool, optional
            Specifies if wind data is given in apparent wind.

            If `True`, data will be converted to true wind.

            Defaults to `False`.
        """
        if apparent_wind:
            new_pts = convert_apparent_wind_to_true(new_pts)
        else:
            new_pts = np.asarray_chkfinite(new_pts)
            if np.any((new_pts[:, 0] <= 0)):
                raise PolarDiagramException(
                    "`new_pts` has non-positive wind speeds"
                )
            new_pts[:, 1] %= 360

        self._points = np.row_stack((self._points, new_pts))

    # TODO Add positivity checks for ws in various cases
    def get_slices(self, ws, n_steps=None, range_=1):
        """For given wind speeds, return the slices of the polar diagram
        corresponding to them.

        The slices then consist of all points in the point cloud where the
        wind speed lies in certain intervals determined by `ws` as below.

        Parameters
        ----------
        ws : See below, optional
            Slices of the polar diagram given as either:

            - a tuple of 2 int/float values, which will be turned into the
            iterable `numpy.linspace(ws[0], ws[1], n_steps)` of int/float
            values.
            The iterable will then be interpreted as below,
            - a mixed iterable containing tuples of 2 int/float values or
            singular int/float values which will be interpreted as
            individual slices. For a tuple the corresponding interval is given
            by the two values of the tuple interpreted as a lower and an upper
            bound. For a singular int/float value `w` the corresponding
            interval will be `(w - range_, w + range_)`,
            - a singular int/float value `w`. The corresponding interval will
            be `(w - range_, w + range_)`.

            If nothing is passed, it will default to
            `(min(self.wind_speeds), max(self.wind_speeds))`.

        n_steps : positive int, optional
            Specifies the amount of slices taken from the given
            interval in `ws`.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed it will default to `int(round(ws[1] - ws[0]))`.

        range_ : positive int or float, optional
            Used to convert an int or float `w` in `ws` to the interval
            `(w - range_, w + range_)`.

            Will only be used if `ws` is int or float or
            if any `w` in `ws` is an int or float.

            Defaults to `1`.

        Returns
        -------
        ws : list
            The wind speeds corresponding to the slices

        wa : numpy.ndarray
            `wa[i]` contains the respective wind angles for wind speed `ws[i]`

        bsp : list of numpy.ndarray
            `bsp[i][j]` contains the resulting boat speed for wind speed
            `ws[i]` and wind angle `wa[i][j]`

        Raises
        ------
        PolarDiagramException
            If `n_steps` is non-positive.
        PolarDiagramException
            If `range_` is non-positive.
        """
        if ws is None:
            ws = self.wind_speeds

        if isinstance(ws, (int, float)):
            ws = [ws]
        elif (
            isinstance(ws, tuple)
            and len(ws) == 2
            and all(isinstance(w, (int, float)) for w in ws)
        ):
            if n_steps is None:
                n_steps = int(round(ws[1] - ws[0]))

            if n_steps <= 0:
                raise PolarDiagramException("`n_steps` is non-positive")

            ws = np.linspace(ws[0], ws[1], n_steps)

        if range_ <= 0:
            raise PolarDiagramException("`range_` is non-positive")

        wa, bsp = self._get_points(ws, range_)

        ws = [(w[0] + w[1]) / 2 if isinstance(w, tuple) else w for w in ws]
        if len(ws) != len(set(ws)):
            warnings.warn(
                "there are duplicate slices. This might cause "
                "unwanted behaviour"
            )

        return ws, wa, bsp

    def _get_points(self, ws, range_):
        wa = []
        bsp = []
        cloud = self.points
        for w in ws:
            if not isinstance(w, tuple):
                w = (w - range_, w + range_)

            pts = cloud[
                np.logical_and(w[1] >= cloud[:, 0], cloud[:, 0] >= w[0])
            ][:, 1:]
            if not pts.size:
                raise PolarDiagramException(
                    f"no points with wind speed in range {w} found"
                )

            # sort for wind angles (needed for plotting methods)
            pts = pts[pts[:, 0].argsort()]

            wa.append(np.deg2rad(pts[:, 0]))
            bsp.append(pts[:, 1])

        if not wa:
            raise PolarDiagramException(
                "there are no slices in the given range `ws`"
            )

        return wa, bsp

    # pylint: disable=arguments-renamed
    def plot_polar(
        self,
        ws=None,
        n_steps=None,
        range_=1,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a polar plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : See below, optional
            Slices of the polar diagram given as either:

            - a tuple of 2 int/float values, which will be turned into the
            iterable `numpy.linspace(ws[0], ws[1], n_steps)` of int/float
            values.
            The iterable will then be interpreted as below,
            - a mixed iterable containing tuples of 2 int/float values or
            singular int/float values which will be interpreted as
            individual slices. For a tuple the corresponding interval is given
            by the two values of the tuple interpreted as a lower and an upper
            bound. For a singular int/float value `w` the corresponding
            interval will be `(w - range_, w + range_)`,
            - a singular int/float value `w`. The corresponding interval will
            be `(w - range_, w + range_)`.

            If nothing is passed, it will default to
            `(min(self.wind_speeds), max(self.wind_speeds))`.

        n_steps : positive int, optional
            Specifies the amount of slices taken from the given
            interval in `ws`.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed it will default to `int(round(ws[1] - ws[0]))`.

        range_ : positive scalar, optional
            Used to convert a scalar `w` in `ws` to the interval
            `(w - range_, w + range_)`.

            Will only be used if `ws` is scalar or
            if any `w` in `ws` is a scalar.

            Defaults to `1`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the legend.

            See `matplotlib.colorbar.Colorbar` and `matplotlib.legend.Legend`
            for possible keywords and their effects.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If `ws` is given as a single value or a list and there is a
            value `w` in `ws`, such that there are no rows in `self.points`
            whose first entry `w` is in the interval `(w-range_, w+range)`.
        """
        ws, wa, bsp = self.get_slices(ws, n_steps, range_)
        plot_polar(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=False,
            **plot_kw,
        )

    # pylint: disable=arguments-renamed
    def plot_flat(
        self,
        ws=None,
        n_steps=None,
        range_=1,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a cartesian plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : See below, optional
            Slices of the polar diagram given as either:

            - a tuple of 2 int/float values, which will be turned into the
            iterable `numpy.linspace(ws[0], ws[1], n_steps)` of int/float
            values.
            The iterable will then be interpreted as below,
            - a mixed iterable containing tuples of 2 int/float values or
            singular int/float values which will be interpreted as
            individual slices. For a tuple the corresponding interval is given
            by the two values of the tuple interpreted as a lower and an upper
            bound. For a singular int/float value `w` the corresponding
            interval will be `(w - range_, w + range_)`,
            - a singular int/float value `w`. The corresponding interval will
            be `(w - range_, w + range_)`.

            If nothing is passed, it will default to
            `(min(self.wind_speeds), max(self.wind_speeds))`.

        n_steps : positive int, optional
            Specifies the amount of slices taken from the given
            interval in `ws`.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed it will default to `int(round(ws[1] - ws[0]))`.

        range_ : positive scalar, optional
            Used to convert a scalar `w` in `ws` to the interval
            `(w - range_, w + range_)`.

            Will only be used if `ws` is scalar or
            if any `w` in `ws` is a scalar.

            Defaults to `1`.

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the legend.

            See `matplotlib.colorbar.Colorbar` and `matplotlib.legend.Legend`
            for possible keywords and their effects.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If `ws` is given as a single value or a list and there is a
            value `w` in `ws`, such that there are no rows in `self.points`
            whose first entry `w` is in the interval `(w-range_, w+range)`.
        """
        ws, wa, bsp = self.get_slices(ws, n_steps, range_)
        wa = [np.rad2deg(a) for a in wa]
        plot_flat(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=False,
            **plot_kw,
        )

    def plot_3d(self, ax=None, colors=("green", "red"), **plot_kw):
        """Creates a 3d plot of the polar diagram.

        Parameters
        ----------
        ax : mpl_toolkits.mplot3d.axes3d.Axes3D, optional
            Axes instance where the plot will be created.

        colors: tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding wind speeds.

            Defaults to `("green", "red")`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If there are no points in the point cloud.
        """
        if not self.points.size:
            raise PolarDiagramException(
                "can't create 3d plot of empty point cloud"
            )

        ws, wa, bsp = (self.points[:, 0], self.points[:, 1], self.points[:, 2])

        wa = np.deg2rad(wa)
        bsp, wa = bsp * np.cos(wa), bsp * np.sin(wa)
        plot3d(ws, wa, bsp, ax, colors, **plot_kw)

    def plot_color_gradient(
        self,
        ax=None,
        colors=("green", "red"),
        marker=None,
        ms=None,
        show_legend=False,
        **legend_kw,
    ):
        """Creates a 'wind speed vs. wind angle' color gradient plot
        of the polar diagram with respect to the corresponding boat speeds.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding boat speed.

            Defaults to `("green", "red")`.

        marker : matplotlib.markers.Markerstyle or equivalent, optional
            Markerstyle for the created scatter plot.

            Defaults to `"o"`.

        ms : float or array_like of fitting shape, optional
            Marker size in points**2.

        show_legend : bool, optional
            Specifies whether a legend will be shown next
            to the plot.

            Legend will be a `matplotlib.colorbar.Colorbar` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        Raises
        ------
        PolarDiagramException
            If there are no points in the point cloud.
        """
        if not self.points.size:
            raise PolarDiagramException(
                "can't create color gradient plot of empty point cloud"
            )

        ws, wa, bsp = (self.points[:, 0], self.points[:, 1], self.points[:, 2])

        plot_color_gradient(
            ws, wa, bsp, ax, colors, marker, ms, show_legend, **legend_kw
        )

    # pylint: disable=arguments-renamed
    def plot_convex_hull(
        self,
        ws=None,
        n_steps=None,
        range_=1,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Computes the (separate) convex hull of one or more
        slices of the polar diagram and creates a polar plot of them.

        Parameters
        ----------
        ws : See below, optional
            Slices of the polar diagram given as either:

            - a tuple of 2 int/float values, which will be turned into the
            iterable `numpy.linspace(ws[0], ws[1], n_steps)` of int/float
            values.
            The iterable will then be interpreted as below,
            - a mixed iterable containing tuples of 2 int/float values or
            singular int/float values which will be interpreted as
            individual slices. For a tuple the corresponding interval is given
            by the two values of the tuple interpreted as a lower and an upper
            bound. For a singular int/float value `w` the corresponding
            interval will be `(w - range_, w + range_)`,
            - a singular int/float value `w`. The corresponding interval will
            be `(w - range_, w + range_)`.

            If nothing is passed, it will default to
            `(min(self.wind_speeds), max(self.wind_speeds))`.

        n_steps : positive int, optional
            Specifies the amount of slices taken from the given
            interval in `ws`.

            Will only be used if `ws` is a tuple of length 2.

            Defaults to `int(round(ws[1] - ws[0]))`.

        range_ : positive int or float, optional

            Will only be used if `ws` is int or float or
            if any `w` in `ws` is an int or float.

            Defaults to `1`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend`.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If `ws` is given as a single value or a list and there is a
            value `w` in `ws`, such that there are no rows in `self.points`
            whose first entry `w` is in the interval `(w-range_, w+range)`.
        """
        ws, wa, bsp = self.get_slices(ws, n_steps, range_)

        plot_convex_hull(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )



# pylint: disable=missing-docstring

import numpy as np


def polynomial_function(x, *params, deg=1):
    """If `deg=n`, models the function
    \\[
        a_0 + a_1x + \\dots + a_nx^n
    \\]
    """
    val = 0
    for i in range(deg + 1):
        val += params[i] * np.power(x, i)

    return val


def inverted_parabola(x, *params):
    """Models the function
    \\[
        a_1 + a_2(x - a_0)^2
    \\]
    """
    return params[1] + params[2] * np.square(x - params[0])


def inverted_shifted_parabola(x, *params):
    """Models the function
    \\[
        a_0x + a_2(x - a_1)^2
    \\]
    """
    return params[0] * x + params[2] * np.square(x - params[1])


def concave_function(x, *params, downturn=True, sat_limit=False):
    """If `downturn` and `sat_limit`, models the function
    \\[
        a_2 - a_0e^{a_1 - x} - a_3x^2
    \\]
    If just `sat_limit`, models the function
    \\[
        a_2 - a_0e^{a_1 - x}
    \\]
    If just `downturn`, models the function
    \\[
        a_0 + a_1x - a_2x^2
    \\]
    """
    if sat_limit and downturn:
        return (
            params[2]
            - params[0] * np.exp(params[1] - x)
            - params[3] * np.square(x)
        )

    if sat_limit:
        return params[2] - params[0] * np.exp(params[1] - x)

    if downturn:
        return polynomial_function(x, params[0], params[1], -params[2], deg=2)

    raise ValueError(
        "At least one of `downturn` and `sat_limit` must be `True`"
    )


def s_shaped(x, *params, downturn=True):
    val = params[2] / (1 + np.exp(params[0] - params[1] * x))
    if downturn:
        val -= params[3] * np.square(x)

    return val


def gompertz_model(x, *params, neg=False):
    if neg:
        return params[2] * np.exp(-params[0] * np.power(-params[1], x))

    return params[2] * np.exp(-params[0] * np.power(params[1], x))


def gaussian_model(x, *params, offset=False):
    val = params[0] * np.exp(-0.5 * np.square(x - params[1]) / params[2])
    if offset:
        val += params[3]

    return val


def gmm_model(x, *params):
    return (
        params[0] * np.exp(-0.5 * np.square(x - params[1]) / params[2])
        + params[3] * np.exp(-0.5 * np.square(x - params[4]) / params[5])
        + params[6]
    )



"""Collection of model functions that can be used with the `Regressor` class
to model various different ship behaviours.

"""


import numpy as np

from ._models2d import (
    concave_function,
    gaussian_model,
    inverted_shifted_parabola,
    s_shaped,
)


def ws_times_wa(ws, wa, scal):
    """
    Returns a scaled version of the product of wind speeds and wind angles.

    Parameters
    --------
    ws: array_like
        Wind speeds.

    wa: array_like
        Wind angles.

    scal: int or float
        The scaling factor.
    """
    ws = np.asarray(ws)
    wa = np.asarray(wa)
    return scal * ws * wa


def ws_concave_dt_wa(ws, wa, *params):
    """A concave function with downturn for wind speed and angle.

    Notes
    -----
    Models the function:

    .. math:: a_0 + a_1x - a_2x^2 + a_3y + a_5(y - a_4)^2 + a_6xy \
    + a_7(360 - y) + a_9((360 - y) - a_8)^2 + a_{10}x(360 - y)

    """
    wa = np.asarray(wa)

    return (
        concave_function(ws, params[0], params[1], params[2])
        + inverted_shifted_parabola(wa, params[3], params[4], params[5])
        + ws_times_wa(ws, wa, params[6])
        + inverted_shifted_parabola(360 - wa, params[7], params[8], params[9])
        + ws_times_wa(ws, 360 - wa, params[10])
    )


def ws_wa_s_dt(ws, wa, *params):
    """An s-shaped function with downturn for wind speed and angle.

    Notes
    -----
    Models the function:

    .. math:: \\frac{a_2}{1 + \\exp(a_0 - a_1x)} - a_3x^2 \
    + \\frac{a_6}{1 + \\exp(a_4 - a_5y)} - a_7y^2 + a_8xy \
    + \\frac{a_{11}}{1 + \\exp(a_9 - a_{10}(360 - y))} - a_{12}(360 - y)^2 \
    + a_{13}x(360 - y)

    """
    wa = np.asarray(wa)

    return (
        s_shaped(ws, params[0], params[1], params[2], params[3])
        + s_shaped(wa, params[4], params[5], params[6], params[7])
        + ws_times_wa(ws, wa, params[8])
        + s_shaped(360 - wa, params[9], params[10], params[11], params[12])
        + ws_times_wa(ws, 360 - wa, params[13])
    )


def ws_s_dt_wa_gauss(ws, wa, *params):
    """An s-shaped function with downturn for wind speed and
    a gaussian function for wind angle.

    Notes
    -----
    Models the function:

    .. math:: \\frac{a_2}{1 + \\exp(a_0 - a_1x)} - a_3x^2 \
    + a_4\\exp\\left(\\frac{-(y - a_5)^2}{2a_6}\\right) \
    + a_7\\exp\\left(\\frac{-((360 - y) - a_8)^2}{2a_9}\\right)

    """
    wa = np.asarray(wa)

    return (
        s_shaped(ws, params[0], params[1], params[2], params[3])
        + gaussian_model(wa, params[4], params[5], params[6])
        + gaussian_model(360 - wa, params[7], params[8], params[9])
    )


def ws_s_s_dt_wa_gauss_comb(ws, wa, *params):
    """

    Notes
    -----
    Models the function:

    .. math:: \\frac{a_2}{1 + \\exp(a_0 - a_1x)} - a_3x^2 \
    + a_4\\exp\\left(\\frac{-(y - a_5)^2}{2a_6}\\right) + a_7xy \
    + a_8\\exp\\left(\\frac{-((360 - y) - a_9)^2}{2a_{10}}\\right) \
    + a_11x(360 - y)

    """
    wa = np.asarray(wa)

    return (
        s_shaped(ws, params[0], params[1], params[2], params[3])
        + gaussian_model(wa, params[4], params[5], params[6])
        + ws_times_wa(ws, wa, params[7])
        + gaussian_model(360 - wa, params[8], params[9], params[10])
        + ws_times_wa(ws, 360 - wa, params[11])
    )


def ws_s_wa_gauss(ws, wa, *params):
    """An s-shaped function for wind speed and a gaussian function for wind
    angle.

    Notes
    -----
    Models the function:

    .. math:: \\frac{a_2}{1 + \\exp(a_0 - a_1x)} \
    + a_3\\exp\\left(\\frac{-(y - a_4)^2}{2a_5}\\right) \
    + a_6\\exp\\left(\\frac{-((360 - y) - a_7)^2}{2a_8}\\right)

    """
    wa = np.asarray(wa)

    return (
        s_shaped(ws, *params[0:4])
        + gaussian_model(wa, *params[4:8])
        + gaussian_model(360 - wa, *params[8:12])
    )


def ws_s_wa_gauss_and_square(ws, wa, *params):
    """

    Notes
    -----
    Models the function

    .. math:: \\left(\\frac{a_2}{1 + \\exp(a_0 - a_1x)} - a_3x^2 \
    + x\\left(a_4\\exp\\left(\\frac{-(y - a_5)^2}{2a_6}\\right) \
    + a_7\\exp\\left(\\frac{-((360 - y) - a_8)^2}{2a_9}\\right) \
    \\right)\\right)y(360 - y)

    """
    ws = np.asarray(ws)
    wa = np.asarray(wa)
    return (
        (
            s_shaped(ws, params[0], params[1], params[2], params[3])
            + ws
            * (
                gaussian_model(wa, params[4], params[5], params[6])
                + gaussian_model(360 - wa, params[7], params[8], params[9])
            )
        )
        * wa
        * (360 - wa)
    )



"""The hrosailing package is a Python library that provides assistance for the
scientific aspects of sailing.
Currently, it implements the data processing framework described
[here](https://www.mdpi.com/2076-3417/12/6/3085) and will be extended in the
future.

In particular, hrosailing provides:

- four kinds of representations of polar diagrams,
- serialization and visualization of polar diagrams,
- creation of polar diagrams from measurement data using
a modular pipeline model and suitable data science methods,
- wind conversion,
- suggestion of optimal tacks and jibes,
- calculation of costs for a sailing trip with respect to the
weather on the way,
- calculation of isochrone points with respect to the weather along the way.

Note that all methods and classes of this package
expect all distances in nautical miles, all speed
values in knots (nautical miles per hour) and all time values in hours
(unless stated otherwise in the respective documentation).

Installation
------------
The recommended way to install `hrosailing` is with
[pip](http://pypi.python.org/pypi/pip).

    pip install hrosailing

[![PyPi version](https://badge.fury.io/py/hrosailing.svg)](https://badge.\
        fury.io/py/hrosailing)

License
-------
The `hrosailing` package is published under the [Apache 2.0 License](https://\
        choosealicense.com/licenses/apache-2.0/).
"""

# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=unused-import

from ._version import __version__ as version

# Tell users if and which hard dependencies are missing
hard_dependencies = ("numpy", "matplotlib", "scipy")
# soft_dependencies = ("pandas", "pynmea2")
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as ie:
        missing_dependencies.append(f"{dependency}: {ie}")

if missing_dependencies:
    raise ImportError(
        "Unable to import required dependencies:\n"
        + "\n".join(missing_dependencies)
    )
del hard_dependencies, dependency, missing_dependencies


import hrosailing.cruising
import hrosailing.pipeline
import hrosailing.pipelinecomponents
import hrosailing.polardiagram
import hrosailing.wind

__all__ = [
    "cruising",
    "pipeline",
    "pipelinecomponents",
    "polardiagram",
    "wind",
    "version",
]



# pylint: disable=missing-module-docstring

import csv
import itertools
import warnings
from ast import literal_eval

import matplotlib.pyplot as plt
import numpy as np

from ._basepolardiagram import (
    PolarDiagram,
    PolarDiagramException,
    PolarDiagramInitializationException,
)
from ._plotting import _get_new_axis, plot_convex_hull_multisails
from ._polardiagramtable import PolarDiagramTable


class NotYetImplementedWarning(Warning):
    """Simple warning for not fully finished implementations."""


class PolarDiagramMultiSails(PolarDiagram):
    """A class to represent, visualize and work with
    a polar diagram made up of multiple sets of sails,
    represented by a `PolarDiagramTable`.


    Class methods aren't fully developed yet. Take care when
    using this class.

    Parameters
    ----------
    pds : Sequence of PolarDiagramTable objects
        Polar diagrams belonging to different sets of sails,
        given as tables, that share the same wind speeds.

    sails : Sequence, optional
        Custom names for the sails. Length should be equal to `pds`.
        If it is not equal it will either be cut off at the appropriate
        length or will be addended with `"Sail i"` to the appropriate length.

        Only important for the legend of plots or the `to_csv()`-method.

        If nothing is passed, the names will be `"Sail i"`, i = 0...n-1,
        where `len(pds) = n`.

    Raises
    ------
    PolarDiagramInitializationException
        If the polar tables don't share the same wind speeds.
    """

    def __init__(self, pds, sails=None):
        warnings.warn(
            (
                "class features aren't all fully developed yet and/or might "
                "change behaviour heavily in the future. "
                "Take care when using this class"
            ),
            category=NotYetImplementedWarning,
        )
        ws = pds[0].wind_speeds
        for pd in pds:
            if not np.array_equal(ws, pd.wind_speeds):
                raise PolarDiagramInitializationException(
                    "wind speed resolution of `pds` does not coincide"
                )

        if sails is None:
            sails = [f"Sail {i}" for i in range(len(pds))]
        elif len(sails) < len(pds):
            sails = list(sails) + [
                f"Sail {i}" for i in range(len(sails) + 1, len(pds))
            ]
        elif len(sails) > len(pds):
            sails = list(sails)
            sails = sails[: len(pds)]

        self._sails = list(sails)
        self._tables = list(pds)

    @property
    def sails(self):
        return self._sails

    @property
    def wind_speeds(self):
        return self._tables[0].wind_speeds

    @property
    def tables(self):
        return self._tables

    def __getitem__(self, item) -> PolarDiagramTable:
        """"""
        try:
            index = self.sails.index(item)
        except ValueError as ve:
            raise PolarDiagramException(
                "`item` is not a name of a sail"
            ) from ve

        return self.tables[index]

    def __str__(self):
        tables = [str(pd) for pd in self._tables]
        names = [str(sail) for sail in self._sails]
        out = []
        for name, table in zip(names, tables):
            out.append(name)
            out.append("\n")
            out.append(table)
            out.append("\n\n")

        return "".join(out)

    def __repr__(self):
        return f"PolarDiagramMultiSails({self.tables}, {self.sails})"

    def to_csv(self, csv_path):
        """Creates a .csv file with delimiter ',' and the
        following format:

            `PolarDiagramMultiSails`
            TWS
            `self.wind_speeds`
            [Sail
            TWA
            `table.wind_angles`
            Boat speeds
            `table.boat_speeds`]

        Parameters
        ----------
        csv_path : path_like
            Path to a .csv file or where a new .csv file will be created.
        """
        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.writer(file, delimiter=",")
            csv_writer.writerow([self.__class__.__name__])
            csv_writer.writerow(["TWS"])
            csv_writer.writerow(self.wind_speeds)
            for sail, table in zip(self.sails, self.tables):
                csv_writer.writerow([sail])
                csv_writer.writerow(["TWA"])
                csv_writer.writerow(table.wind_angles)
                csv_writer.writerow(["BSP"])
                csv_writer.writerows(table.boat_speeds)

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=",")
        next(csv_reader)
        ws_resolution = [literal_eval(ws) for ws in next(csv_reader)]
        sails, pds = _extract_polardiagrams(csv_reader, ws_resolution)
        return PolarDiagramMultiSails(pds, sails)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram, by
        mirroring each `PolarDiagramTable` at the 0° - 180° axis and
        returning a new instance. See also the `symmetrize()`-method
        of the `PolarDiagramTable` class.

        Warning
        -------
        Should only be used if all the wind angles of the `PolarDiagramTables`
        are each on one side of the 0° - 180° axis, otherwise this can lead
        to duplicate data, which can overwrite or live alongside old data.
        """
        pds = [pd.symmetrize() for pd in self._tables]
        return PolarDiagramMultiSails(pds, self._sails)

    def get_slices(self, ws):
        """For given wind speeds, return the slices of the polar diagram
        corresponding to them.

        The slices are equal to the corresponding
        columns of the table together with `self.wind_angles`.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram table, given as either:

            - a tuple of length 2 specifying an interval of considered
            wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

            If nothing it passed, it will default to `self.wind_speeds`.


        Returns
        -------
        ws : list
            The wind speeds corresponding to the slices.

        wa : list of numpy.ndarray
            A list of the corresponding wind angles for each slice.

        bsp : list of numpy.ndarray
            `bsp[i][j]` contains the resulting boat speed for wind speed
            `ws[i]` and wind angle `wa[i][j]`.

        members : list of str
            `members[j]` contains the name of the sail corresponding to the
            wind `ws[i]` and `wa[i][j]` for any value of `i`.
        """
        wa = []
        temp = []
        for pd in self._tables:
            ws, w, b = pd.get_slices(ws)
            wa.append(w)
            temp.append(b)

        flatten = itertools.chain.from_iterable
        members = [[self._sails[i]] * len(w) for i, w in enumerate(wa)]
        members = list(flatten(members))

        wa = [np.asarray(wa).ravel()] * len(ws)
        bsp = []
        for i in range(len(ws)):
            b = np.asarray([b_[:, i] for b_ in temp]).ravel()
            bsp.append(b)

        return ws, wa, bsp, members

    def plot_polar(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a polar plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram table, given as either:

            - a tuple of length 2 specifying an interval of considered
            wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

            The slices are then equal to the corresponding
            columns of the table together with `self.wind_angles`.

            If nothing it passed, it will default to `self.wind_speeds`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend`.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        PolarDiagramException
            If the given interval doesn't contain any slices of the
            polar diagram.
        """
        if ax is None:
            ax = plt.axes(projection="polar")

        for i, pd in enumerate(self._tables):
            if i == 0 and show_legend:
                pd.plot_polar(
                    ws, ax, colors, show_legend, legend_kw, **plot_kw
                )
                continue

            pd.plot_polar(ws, ax, colors, False, None, **plot_kw)

    def plot_flat(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a cartesian plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram table, given as either:

            - a tuple of length 2 specifying an interval of considered
            wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

            The slices are then equal to the corresponding
            columns of the table together with `self.wind_angles`.

            If nothing is passed, it will default to `self.wind_speeds`.

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether or not a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend`.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        PolarDiagramException
            If the given interval doesn't contain any slices of the
            polar diagram.
        """
        if ax is None:
            ax = _get_new_axis("rectilinear")

        for i, pd in enumerate(self._tables):
            if i == 0 and show_legend:
                pd.plot_flat(ws, ax, colors, show_legend, legend_kw, **plot_kw)
                continue

            pd.plot_flat(ws, ax, colors, False, None, **plot_kw)

    def plot_3d(self, ax=None, colors=("green", "red")):
        """Creates a 3d plot of the polar diagram.

        Parameters
        ----------
        ax : mpl_toolkits.mplot3d.axes3d.Axes3D, optional
            Axes instance where the plot will be created.

        colors: tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding wind speeds.

            Defaults to `("green", "red")`.
        """
        if ax is None:
            ax = plt.axes(projection="3d")

        for pd in self._tables:
            pd.plot_3d(ax, colors)

    def plot_color_gradient(
        self,
        ax=None,
        colors=("green", "red"),
        marker=None,
        ms=None,
        show_legend=False,
        **legend_kw,
    ):
        """

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding boat speed.

            Defaults to `("green", "red")`.

        marker : matplotlib.markers.Markerstyle or equivalent, optional
            Markerstyle for the created scatter plot.

            Defaults to `"o"`.

        ms : float or array_like of fitting shape, optional
            Marker size in points**2.

        show_legend : bool, optional
            Specifies whether or not a legend will be shown next
            to the plot.

            Legend will be a `matplotlib.colorbar.Colorbar` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        """
        warnings.warn(
            "feature isn't implemented yet", category=NotYetImplementedWarning
        )

    def plot_convex_hull(
        self,
        ws=None,
        ax=None,
        colors=None,
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Computes the (separate) convex hull of one or more
        slices of the polar diagram and creates a polar plot of them.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram table, given as either:

            - a tuple of length 2 specifying an interval of considered
            wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

            The slices are then equal to the corresponding
            columns of the table together with `self.wind_angles`.

            If nothing it passed, it will default to `self.wind_speeds`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : subscriptable iterable of color_likes, optional

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend`.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        """
        ws, wa, bsp, members = self.get_slices(ws)
        plot_convex_hull_multisails(
            ws, wa, bsp, members, ax, colors, show_legend, legend_kw, **plot_kw
        )


def _extract_polardiagrams(csv_reader, ws_resolution):
    sails = []
    pds = []

    while True:
        try:
            sails.append(next(csv_reader)[0])
            next(csv_reader)
            wa_resolution = [literal_eval(wa) for wa in next(csv_reader)]
            next(csv_reader)
            bsps = [
                [literal_eval(bsp) for bsp in row]
                for row in itertools.islice(csv_reader, len(wa_resolution))
            ]
            pds.append(PolarDiagramTable(ws_resolution, wa_resolution, bsps))
        except StopIteration:
            break

    return sails, pds



"""
Components for the `PolarPipeline` and `PipelineExtension`
classes among other things.
"""

from .datahandler import (
    ArrayHandler,
    CsvFileHandler,
    DataHandler,
    NMEAFileHandler,
)
from .expander import Expander, LazyExpander, WeatherExpander
from .filter import BoundFilter, Filter, QuantileFilter
from .imputator import FillLocalImputator, RemoveOnlyImputator
from .influencemodel import (
    IdentityInfluenceModel,
    InfluenceModel,
    WindAngleCorrectingInfluenceModel,
)
from .injector import ZeroInjector
from .interpolator import (
    ArithmeticMeanInterpolator,
    IDWInterpolator,
    ImprovedIDWInterpolator,
    Interpolator,
)
from .neighbourhood import (
    Ball,
    Cuboid,
    Ellipsoid,
    Neighbourhood,
    Polytope,
    ScalingBall,
)
from .quality_assurance import (
    ComformingQualityAssurance,
    MinimalQualityAssurance,
    QualityAssurance,
)
from .regressor import LeastSquareRegressor, ODRegressor, Regressor
from .sampler import (
    ArchimedeanSampler,
    FibonacciSampler,
    Sampler,
    UniformRandomSampler,
)
from .smoother import AffineSmoother, LazySmoother, Smoother
from .weigher import (
    AllOneWeigher,
    CylindricMeanWeigher,
    CylindricMemberWeigher,
    FluctuationWeigher,
    FuzzyVariable,
    FuzzyWeigher,
    Weigher,
    WeightedPoints,
)



"""Contains the class `Data` which is an output of several pipeline components.
"""

from datetime import datetime
from decimal import Decimal

import numpy as np

from hrosailing.pipelinecomponents.constants import KEYSYNONYMS, SEPARATORS


class Data:
    """
    Organizes measurement data. Data is interpreted as a table with columns
    identified by keywords and rows corresponding to a certain measurement
    point. Also saves and checks the data types of the corresponding columns.
    Weights might be added to the corresponding rows.

    If `data` is of type `Data`, `data[key]` yields the corresponding column if
    `key` is a string and the corresponding row if `key` is an integer.
    `key in data` checks if there is a column with key `key`.
    Iteration is performed over the rows.

    Properties
    ----------
    numerical : numpy.ndarray
        An array containing all data of type `float`.

    n_rows : int
        The number of rows.

    n_cols : int
        The number of columns.
    """

    def __init__(self):
        self._data = {}
        self._types = {}
        self._max_len = 0

    def keys(self):
        """
        Returns
        -------
        keys : list of str
            List of all column names.
        """
        return list(self._data.keys())

    @property
    def numerical(self):
        float_keys, float_vals = self.get_by_type(float)
        array = np.column_stack(float_vals)
        return float_keys, array

    @property
    def n_rows(self):
        return self._max_len

    @property
    def n_cols(self):
        return len(self._data)

    @property
    def data(self):
        return self._data.copy()

    def rows(self, keys=None, return_type=dict):
        """
        Iterates over the rows.

        Parameters
        ----------
        keys : iterable over str, optional
            The columns from which the iterator yields. Other columns will be ignored.

            Defaults to `self.keys()`.

        return_type : dict, tuple or list
            Defines the type of the yielded data. If types `tuple` or `list` are used, the entries are in the exact same
            order as in `keys`.

            Defaults to `dict`.
        """
        if keys is None:
            keys = self.keys()
        for i in range(self._max_len):
            if return_type is dict:
                yield {key: self._data[key][i] for key in keys}
            elif return_type is tuple:
                yield (self._data[key][i] for key in keys)
            elif return_type is list:
                yield [self._data[key][i] for key in keys]
            else:
                raise ValueError(
                    f"Return type of {return_type} is not supported"
                )

    def get_by_type(self, type_):
        """
        Parameters
        ----------
        type_ : type

        Returns
        -------
        filtered_keys : list of str
            A list of all keys with columns of type `type_`.

        filtered_data : list of lists
            A list of all data corresponding to the columns with keys
            `filtered_keys`.
        """
        filtered_keys = [
            key for key in self.keys() if self._types[key] == type_
        ]
        return filtered_keys, [self._data[key] for key in filtered_keys]

    def type(self, key):
        """
        Returns the associated type corresponding to the `key` column.

        Parameters
        ----------
        key : str
        """
        return self._types[key]

    def extend(self, key, data):
        """
        Extends the data by given other data.

        Parameters
        ----------
        key : str
            The key of the column that should be extended by the given data.
        data : list
            Data that should be appended to the given column.
        """
        data_type = self._get_type(data)
        if key in self._data:
            if data_type is not None and data_type != self._types[key]:
                raise ValueError(
                    f"data should be of type {self._types[key]}"
                    f" but has type {data_type}"
                )
            self._data[key].extend(data)
            self._max_len = max(self._max_len, len(self._data[key]))
        else:
            self.fill(self._max_len - len(data), [key])
            self._data[key].extend(data)
            self._types[key] = data_type
            self._max_len = max(self._max_len, len(self._data[key]))

    def update(self, data_dict):
        """
        Extends the data according to given data and fills missing
        entries in each column with `None`.

        Parameters
        ----------
        data_dict : dict or Data
            The dictionary or `Data` object containing the data to be used for the update.
        """
        if isinstance(data_dict, dict):
            for key, val in data_dict.items():
                if isinstance(val, list):
                    self.extend(key, val)
                else:
                    self.append(key, val)
            self.fill()
        # if isinstance(data_dict, Data): this does not work for some reason
        else:
            self.update(data_dict.data)

    def append(self, key, data):
        """
        Extends the data by a single element.

        Parameters
        ----------
        key : str
            The key of the column to which the given data should be appended.
        data : object
            Single data object to be appended.

        See also
        --------
        `Data.extend`
        """
        self.extend(key, [data])

    def fill(self, len_=None, keys=None):
        """
        Fills all columns smaller than a specific length with `None` values.

        Parameters
        ----------
        len_ : int, optional
            The required length.

            Defaults to `self.n_rows`.

        keys : list of str, optional
            The keys of the columns which should be filled.

            Defaults to `self.keys`.
        """
        if keys is None:
            keys = self.keys()
        if len_ is None:
            len_ = self._max_len
        for key in keys:
            if key not in self._data:
                self._data[key] = []
            curr_len = len(self._data[key])
            fill_len = max(0, (len_ - curr_len))
            self._data[key].extend([None] * fill_len)
            self._types[key] = self._get_type(self._data[key])

    def filter_types(self, type_list):
        """
        Deletes all data which is not in a list of required types.

        Parameters
        ----------
        type_list : list of types
        """
        for key in list(self._data.keys()):
            if self._types[key] not in type_list:
                del self._data[key]
                del self._types[key]

    def rename(self, old_key, new_key):
        """
        Renames a column.

        Parameters
        ----------
        old_key : str
            Old name of column.

        new_key : str
            New name of column.
        """
        if old_key == new_key:
            return
        if new_key in self.keys():
            raise ValueError(
                f"Can not rename {old_key} to {new_key} since"
                f" {new_key} already exists"
            )

        self._data[new_key] = self._data[old_key]
        self._types[new_key] = self._types[old_key]

        self.delete(old_key)

    def delete(self, key):
        """
        Deletes a column or a row.

        Parameters
        ----------
        key : int, str, list of str or list of int
            Name(s) of column(s) to be deleted or index(es) of row(s) to be deleted.
        """
        if isinstance(key, str):
            del self._data[key]
            del self._types[key]
        if isinstance(key, int):
            self.delete([key])
        if isinstance(key, list):
            if len(key) == 0:
                return
            if isinstance(key[0], str):
                for k in key:
                    self.delete(k)
            if isinstance(key[0], int):
                self._data = {
                    k: [v for i, v in enumerate(value) if i not in key]
                    for k, value in self._data.items()
                }
                self._max_len -= len(key)

    def strip(self, mode="all"):
        """
        Removes either columns that only consist of `None` values or leading and tailing rows consisting only of `None`
        values.

        Parameters
        ----------
        mode : {"cols", "rows", "all"}
            Specifies whether columns or rows or both should be removed.

            Defaults to "all"
        """
        if mode == "cols":
            self._data = {
                key: value
                for key, value in self._data.items()
                if not all(v is None for v in value)
            }
        elif mode == "rows":
            end_leading_nones = 0  # first index AFTER leading None rows
            start_tailing_nones = (
                self._max_len - 1
            )  # first index BEFORE tailing None rows
            while all(
                value[end_leading_nones] is None
                for key, value in self._data.items()
            ):
                end_leading_nones += 1
            while all(
                value[start_tailing_nones] is None
                for key, value in self._data.items()
            ):
                start_tailing_nones -= 1
            self._data = {
                key: value[end_leading_nones : start_tailing_nones + 1]
                for key, value in self._data.items()
            }
        elif mode == "all":
            self.strip("cols")
            self.strip("rows")

    def hrosailing_standard_format(self):
        """
        Reformats data into the `hrosailing` standard format.

        This means:
            - the dictionary has `hrosailing` standard keys whenever possible,
            - date and time fields will be aggregated to datetime,
            - tries to cast entries to `float` whenever possible.
        """

        def get_standard_key(key):
            lkey = key.lower()
            for sep in SEPARATORS:
                lkey = lkey.replace(sep, " ")
            lkey = lkey.strip()
            # remove plural
            stripped_lkey = lkey if lkey[-1] != "s" else lkey[:-1]
            if lkey in KEYSYNONYMS:
                return KEYSYNONYMS[lkey]
            if stripped_lkey in KEYSYNONYMS:
                return KEYSYNONYMS[stripped_lkey]

            return key

        for key, value in list(self._data.items()):
            standard_key = get_standard_key(key)
            new_key = standard_key
            if key == new_key:
                continue
            i = 0
            while new_key in self.keys():
                i += 1
                new_key = f"{standard_key}(Control value {i})"
            self.rename(key, new_key)

        if "time" in self and "date" in self:

            def combine(date, time):
                if date is None or time is None:
                    return None
                return datetime.combine(date, time)

            self.extend(
                "datetime",
                [
                    combine(date, time)
                    for date, time in zip(self["date"], self["time"])
                ],
            )
            self.delete("date")
            self.delete("time")

        # ensure floats
        for key in self.keys():
            if self._types[key] in (int, str, Decimal, np.float_):
                succes, self._data[key] = _try_call_to_float(self._data[key])
                if succes:
                    self._types[key] = float

    @staticmethod
    def _get_type(data):
        curr_type = None
        for entry in data:
            if entry is None:
                continue
            if curr_type is None:
                curr_type = type(entry)
                continue
            if type(entry) != curr_type:
                raise ValueError(
                    "Data has no consistent type."
                    f"Found the types {type(entry)} and {curr_type}"
                )
        return curr_type

    @classmethod
    def concatenate(cls, list_):
        """
        Returns concatenated data, given a list of `Data` classes.

        Parameters
        ----------
        list_ : list of Data

        Returns
        -------
        data : Data
            The concatenated Data.
        """
        data = cls()
        for other_data in list_:
            data.update(other_data)
        return data

    @classmethod
    def _force_set(cls, data, types, max_len):
        # use with caution: wrong usage can produce inconsistent `Data` objects.
        new_obj = cls()
        new_obj._data = data
        new_obj._types = types
        new_obj._max_len = max_len
        return new_obj

    @classmethod
    def from_dict(cls, dict_):
        """
        Creates a `Data` object that contains the same data as a given dictionary.
        The keys and values of the dictionary should be iterables and will correspond to the columns of the resulting
        `Data` object.

        Parameters
        ----------
        dict_ : dict
        """
        data = cls()
        data.update(dict_)
        return data

    def get_slice(self, slice_):
        """
        Returns new `Data` object containing only keys in `slice`.

        Parameters
        ----------
        slice_ : object with method `__contains__`

        Returns
        -------
        data: Data
        """

        data = {
            key: value for key, value in self._data.items() if key in slice_
        }
        types = {
            key: value for key, value in self._types.items() if key in slice_
        }
        try:
            max_len = max(len(field) for field in data.values())
        except ValueError:
            max_len = 0
        return Data._force_set(data, types, max_len)

    def _mask_rows(self, mask):
        """
        Keeps only a subset of the rows indicated by `mask`.

        Parameters
        ----------
        mask : array_like of booleans
        """

        data = {
            key: [entry for entry, choose in zip(value, mask) if choose]
            for key, value in self._data.items()
        }
        try:
            max_len = max(len(field) for field in data.values())
        except ValueError:
            max_len = 0

        return Data._force_set(data, self._types.copy(), max_len)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._data[item]
        if isinstance(item, int):
            return {key: val[item] for key, val in self._data.items()}
        type_ = None
        for key in item:
            type_ = type(key)
            break
        if type_ is None or type_ in (bool, np.bool_):
            return self._mask_rows(item)
        if type_ is str:
            return self.get_slice(item)
        raise TypeError(
            "Only types `int`, `str` and iterables over `bool`, `numpy.bool_`"
            " or `str` are supported"
        )

    def __contains__(self, item):
        return item in self._data

    def __iter__(self):
        for i in range(self._max_len):
            yield self[i]

    def __str__(self):
        str_ = ""
        for key in self.keys():
            str_ += f"Data field '{key}' of type {self._types[key]}:\n\t"
            if len(self._data[key]) < 10:
                str_ += str(self._data[key])
            else:
                str_ += (
                    f"[{','.join([str(item) for item in self._data[key][:5]])},"
                    " ..."
                    f" ,{','.join([str(item) for item in self._data[key][-5:]])}]"
                )
            str_ += "\n"

        return str_


def _try_call_to_float(list_):
    new_list = []
    for i, value in enumerate(list_):
        if value is None:
            new_list.append(value)
            continue
        try:
            new_list.append(float(value))
        except (TypeError, ValueError):
            return False, list_
    return True, new_list



# pylint: disable=missing-module-docstring
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hrosailing",
    version="0.10.0",
    author="Valentin Dannenberg & Robert Schueler",
    author_email="valentin.dannenberg2@uni-rostock.de",
    description="Python library for Polar (Performance) Diagrams",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hrosailing/hrosailing",
    project_urls={
        "Bug Tracker": "https://github.com/hrosailing/hrosailing/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=[
        "hrosailing",
        "hrosailing.cruising",
        "hrosailing.polardiagram",
        "hrosailing.pipeline",
        "hrosailing.pipelinecomponents",
        "hrosailing.pipelinecomponents.modelfunctions",
    ],
    python_requires=">=3.7",
)



# pylint: disable=missing-module-docstring

from abc import ABC, abstractmethod


class PolarDiagramException(Exception):
    """Exception raised if some nonstandard error occurs,
    while doing something with polar diagrams.
    """


class PolarDiagramInitializationException(Exception):
    """Exception raised if an error occurs during
    initialization of a `PolarDiagram`.
    """


class PolarDiagram(ABC):
    """Base class for all polar diagrams.

    Abstract Methods
    ----------------
    to_csv(csv_path)

    __from_csv__(cls, file)

    symmetrize()

    get_slices(ws)

    plot_polar(
        ws,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw
    )

    plot_flat(
        ws,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw
    )

    plot_3d(ax=None, **plot_kw)

    plot_color_gradient(
        ax=None,
        colors=("green", "red"),
        marker=None,
        ms=None,
        show_legend=False,
        **legend_kw,
    )

    plot_convex_hull(
        ws,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    )
    """

    @abstractmethod
    def to_csv(self, csv_path):
        """This method should, given a path, write a .csv file in
        the location, containing human-readable information about the
        polar diagram object that called the method.

        Parameters
        ----------
        csv_path: str
            Path of the created csv file.
        """

    @classmethod
    @abstractmethod
    def __from_csv__(cls, file):
        """"""

    @abstractmethod
    def symmetrize(self):
        """This method should return a new `PolarDiagram` object that is a
        symmetric (i.e. mirrored along the 0 - 180° axis) version of the
        `PolarDiagram` object called the method.
        """

    @abstractmethod
    def get_slices(self, ws):
        """This method should, given a number of wind speeds, return
        a list of the given wind speeds as well as wind angles and
        corresponding boat speeds, that reflect how the vessel behaves at
        the given wind speeds.

        Parameters
        ----------
        ws : int/float
            Description of slices of the polar diagram to be plotted.

            For a description of what the slice is made of,
            see the `get_slices()`-method of the respective
            `PolarDiagram` subclass.
        """

    def plot_polar_slice(self, ws, ax=None, **plot_kw):
        """Creates a polar plot of a given slice of the
        polar diagram.

        Parameters
        ----------
        ws : int/float
            Description of slices of the polar diagram to be plotted.

            For a description of what the slice is made of,
            see the `plot_polar()`-method of the respective
            `PolarDiagram` subclass.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

            If nothing is passed, the function will create
            a suitable axes.

        plot_kw : Keyword arguments
            Keyword arguments that will be passed to the
            `matplotlib.axes.Axes.plot` function, to change
            certain appearances of the plot.
        """
        self.plot_polar(
            ws, ax, colors=None, show_legend=False, legend_kw=None, **plot_kw
        )

    def plot_flat_slice(self, ws, ax=None, **plot_kw):
        """Creates a cartesian plot of a given slice of the
        polar diagram.

        Parameters
        ----------
        ws : int/float
            Slice of the polar diagram.

            For a description of what the slice is made of,
            see the `plot_flat()`-method of the respective
            `PolarDiagram` subclass.

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

            If nothing is passed, the function will create
            a suitable axes.

        plot_kw : Keyword arguments
            Keyword arguments that will be passed to the
            `matplotlib.axes.Axes.plot` function, to change
            certain appearances of the plot.
        """
        self.plot_flat(
            ws, ax, colors=None, show_legend=False, legend_kw=None, **plot_kw
        )

    @abstractmethod
    def plot_polar(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """This method should create a polar plot of one or more slices,
        corresponding to `ws`, of the polar diagram object.

        Parameters
        ---------

        ws : int/float
            Description of slices of the polar diagram to be plotted.

            For details refer to the `plot_polar()`-method of the respective
            `PolarDiagram` subclass.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            For details refer to the `plot_polar()`-method of the respective
            `PolarDiagram` subclass.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether or not a legend will be shown next to the plot.

            For details refer to the `plot_polar()`-method of the respective
            `PolarDiagram` subclass.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the legend.

            See `matplotlib.colorbar.Colorbar` for possible keywords
            when a colorbar is created and `matplotlib.legend.Legend` for
            possible keywords in cases where a Legend is created.

            For details refer to the `plot_polar()`-method of the respective
            `PolarDiagram` subclass.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.
        """

    @abstractmethod
    def plot_flat(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """This method should create a cartesian plot of one or more slices,
        corresponding to `ws`, of the `PolarDiagram` object.


        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram.

            For details refer to the `plot_flat()`-method of the respective
            `PolarDiagram` subclass.

            If nothing is passed, it will default to `(0, 20)`.

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            For details refer to the `plot_flat()`-method of the respective
            `PolarDiagram` subclass.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            For details refer to the `plot_flat()`-method of the respective
            `PolarDiagram` subclass.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the legend.

            See `matplotlib.colorbar.Colorbar` and `matplotlib.legend.Legend` for
            possible keywords and their effects.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.
        """

    @abstractmethod
    def plot_3d(self):
        """This method should create a 3d plot of the polar diagram object."""

    @abstractmethod
    def plot_color_gradient(
        self,
        ax=None,
        colors=("green", "red"),
        marker=None,
        ms=None,
        show_legend=False,
        **legend_kw,
    ):
        """This method should create 'wind speed vs. wind angle'
        color gradient plot of the polar diagram object with respect
        to the corresponding boat speeds.

        Parameters
        ----------

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            For details refer to the `plot_color_gradient()`-method of
            the respective `PolarDiagram` subclass.

            Defaults to `("green", "red")`.

        marker : matplotlib.markers.Markerstyle or equivalent, optional
            Markerstyle for the created scatter plot.

            Defaults to `"o"`

        ms : float or array_like of fitting shape, optional
            Marker size in points**2.

        show_legend : bool, optional
            Specifies whether or not a legend will be shown next
            to the plot.

            Legend will be a `matplotlib.colorbar.Colorbar` instance.

            Defaults to `False`.

        legend_kw : Keyword arguments
            Keyword arguments to change position and appearance of the legend.

            See `matplotlib.legend.Legend` for possible keywords and
            their effects.

            Will only be used if `show_legend` is `True`.
        """

    def plot_convex_hull_slice(self, ws, ax=None, **plot_kw):
        """Computes the convex hull of a given slice of
        the polar diagram and creates a polar plot of it.

        Parameters
        ----------
        ws : int/float
            Slice of the polar diagram.

            For a description of what the slice is made of,
            see the `plot_convex_hull()`-method of the respective
            `PolarDiagram` subclass.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

            If nothing is passed, the function will create
            a suitable axes.

        plot_kw : Keyword arguments
            Keyword arguments that will be passed to the
            `matplotlib.axes.Axes.plot` function, to change
            certain appearances of the plot.
        """
        self.plot_convex_hull(
            ws, ax, colors=None, show_legend=False, legend_kw=None, **plot_kw
        )

    @abstractmethod
    def plot_convex_hull(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """This method should compute the convex hull of one or multiple
        slices, corresponding to `ws`, of the polar diagram and then create
        a polar plot of them.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram.

            For details refer to the `plot_convex_hull()`-method of
            the respective `PolarDiagram` subclass.

            If nothing is passed, it will default to `(0, 20)`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            For details refer to the `plot_convex_hull()`-method of
            the respective `PolarDiagram` subclass.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether or not a legend will be shown next to the plot.

            The type of legend depends on the color options.

            For details refer to the `plot_convex_hull()`-method of
            the respective `PolarDiagram` subclass.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the legend.

            See `matplotlib.colorbar.Colorbar` and `matplotlib.legend.Legend`
            for possible keywords and their effects.

            Will only be used if show_legend is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        """



"""
Classes used for modular modelling of different interpolation methods.

Defines the `Interpolator` abstract base class that can be used to
create custom interpolation methods.

Subclasses of `Interpolator` can be used with:

- the `TableExtension` and `PointcloudExtension` classes in the
`hrosailing.pipeline` module,
- the `__call__`-methods of the `PolarDiagramTable` and `PolarDiagramPointcloud`
classes in the `hrosailing.polardiagram` module.
"""


from abc import ABC, abstractmethod
from typing import Callable

import numpy as np

from ._utils import scaled_euclidean_norm
from .neighbourhood import Neighbourhood


class InterpolatorInitializationException(Exception):
    """Exception raised if an error occurs during
    initialization of an `Interpolator`.
    """


class Interpolator(ABC):
    """Base class for all `Interpolator` classes.


    Abstract Methods
    ----------------
    interpolate(self, w_pts, grid_pt)
    """

    @abstractmethod
    def interpolate(self, w_pts, grid_pt):
        """This method should be used, given a point `grid_pt` and an
        instance of `WeightedPoints`, to determine the z-value at `grid_pt`,
        based on the z-values of the points in the `WeightedPoints` instance.

        Parameters
        ----------
        w_pts : WeightedPoints
            Considered measured points.

        grid_pt : numpy.ndarray of shape (2,)
            Point that is to be interpolated.

        Returns
        -------
        out : int / float
            Interpolated value at `grid_pt`.
        """


class IDWInterpolator(Interpolator):
    """Basic inverse distance interpolator, based on the work
    of Shepard, "A two-dimensional interpolation function
    for irregularly-spaced data".

    For a given point :math:`grid\\\\_pt`, that is to be interpolated, we
    calculate the distances :math:`d_{pt} = ||grid\\\\_pt - pt[:2]||` for all considered
    measured points. Then we set the weights of a point :math:`pt` to be
    :math:`w_{pt} = \\cfrac{1}{d_{pt}^p}`, for some non-negative integer :math:`p`.

    The interpolated value at :math:`grid\\\\_pt` then equals
    :math:`\\cfrac{\\sum_{pt} w_{pt} * pt[2]}{\\sum_{pt} w_{pt}}`
    or if :math:`grid\\\\_pt` is already a measured point :math:`pt`, it will equal :math:`pt[2]`.

    Parameters
    ----------
    p : non-negative int, optional

        Defaults to `2`.

    norm : function or callable, optional
        Norm with which to calculate the distances, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled
        version of :math:`||.||_2`.

    Raises
    ------
    InterpolatorInitializationException
        If `p` is negative.
    """

    def __init__(self, p=2, norm: Callable = scaled_euclidean_norm):
        super().__init__()
        if p < 0:
            raise InterpolatorInitializationException("`p` is negative")

        self._p = p
        self._norm = norm

    def __repr__(self):
        return f"IDWInterpolator(p={self._p}, norm={self._norm.__name__})"

    def interpolate(self, w_pts, grid_pt):
        """Interpolates a given point `grid_pt` according to the procedure described
        above.

        Parameters
        ----------
        w_pts : WeightedPoints
            Considered measured points.

        grid_pt : numpy.ndarray of shape (2,)
            Point that is to be interpolated.

        Returns
        -------
        out : int / float
            Interpolated value at `grid_pt`.
        """
        pts = w_pts.data
        wts = self._norm(pts[:, :2] - grid_pt)
        if np.any(wts == 0):
            mask = wts == 0
            return pts[mask][0, 2]

        wts = 1 / np.power(wts, self._p)
        wts /= np.sum(wts)

        return wts @ pts[:, 2]


def _gauss_potential(distances, weights, *params):
    alpha = params[0]
    return np.exp(-alpha * weights * distances)


class ArithmeticMeanInterpolator(Interpolator):
    """An interpolator that computes the interpolated value according
    to the following procedure.

    First the distance of the independent variables (wind angle and wind speed)
    of all considered
    points and of the to interpolate point is calculated, i.e.
    :math:`|| p[:2] - grid\\\\_pt ||`.
    Then using a distribution, new weights are calculated based on
    the old weights, the previously calculated distances and other
    parameters depending on the distribution.

    The value of the dependent variable (boat speed) of the interpolated point
    then equals :math:`s * \\cfrac{\\sum_p w_p * p}{\\sum_p w_p}`
    where :math:`s` is an additional scaling factor.

    Note that this is a more general approach to the inverse distance
    interpolator.

    Parameters
    ----------
    s : positive int or float, optional
        Scaling factor for the arithmetic mean.

        Defaults to `1`.

    norm : function or callable, optional
        Norm with which to calculate the distances, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled
        version of :math:`||.||_2`.

    distribution : function or callable, optional
        Function with which to calculate the updated weights.

        Should have the signature
        `f(distances, old_weights, *parameters) -> new_weights`.

        If nothing is passed, it will default to `gauss_potential`, which
        calculates weights based on the formula
        :math:`\\mathrm{e}^{\\textstyle -\\alpha * oldweights * distances}\\` with
        parameter :math:`\\alpha`.

    params : tuple
        Parameters to be passed to `distribution`.

        If no `distribution` is passed it defaults to `(100,)`, otherwise to `()`.

    Raises
    ------
    InterpolatorInitializationException
        If `s` is non-positive.
    """

    def __init__(
        self,
        s=1,
        norm: Callable = scaled_euclidean_norm,
        distribution: Callable = None,
        params=(),
    ):
        super().__init__()
        if s <= 0:
            raise InterpolatorInitializationException("`s` is non-positive")

        self._s = s
        self._norm = norm
        self._params = params
        if distribution is None:
            self._distr = _gauss_potential
            if params == ():
                self._params = (100,)
        else:
            self._distr = distribution

    def __repr__(self):
        return (
            f"ArithmeticMeanInterpolator(s={self._s},"
            f" norm={self._norm.__name__}, distribution={self._distr},"
            f" params={self._params})"
        )

    def interpolate(self, w_pts, grid_pt):
        """Interpolates a given `grid_pt` according to the procedure described
        above.

        Parameters
        ----------
        w_pts : WeightedPoints
            Considered measured points.

        grid_pt : numpy.ndarray of shape (2,)
            Point that is to be interpolated.

        Returns
        -------
        out : int / float
            Interpolated value at `grid_pt`.
        """
        pts = w_pts.data
        dist = self._norm(pts[:, :2] - grid_pt)
        if np.any(dist == 0):
            mask = dist == 0
            return pts[mask][0, 2]

        wts = w_pts.weights
        wts = self._distr(dist, wts, *self._params)
        return self._s * np.average(pts[:, 2], weights=wts)


class ImprovedIDWInterpolator(Interpolator):
    """An improved inverse distance interpolator, based
    on the work of Shepard, "A two-dimensional interpolation
    function for irregularly-spaced data".

    Instead of setting the weights as the normal inverse distance
    to some power, we set the weights in the following way:

    Let :math:`r` be the radius of the `neighbourhood.ScalingBall` with the center being some
    point :math:`grid\\\\_pt` which is to be interpolated.
    For all considered measured points let :math:`d_{pt}` be the same as
    in `IDWInterpolator`. If :math:`d_{pt} \\leq \\cfrac{r}{3}` we set :math:`w_{pt} = \\cfrac{1}{d_{pt}}`.
    Otherwise, we set :math:`w_{pt} = \\cfrac{27}{4 * r} * (\\cfrac{d_{pt}}{r - 1})^2`.

    The resulting value at :math:`grid\\\\_pt` will then be calculated the same
    way as in `IDWInterpolator`.

    Parameters
    ----------
    norm : function or callable, optional
        Norm with which to calculate the distances, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled
        version of :math:`||.||_2`.
    """

    def __init__(self, norm: Callable = scaled_euclidean_norm):
        super().__init__()
        self._norm = norm

    def interpolate(self, w_pts, grid_pt):
        """Interpolates a given `grid_pt` according to the procedure described
        above.

        Parameters
        ----------
        w_pts : WeightedPoints
            Considered measured points.

        grid_pt : numpy.ndarray of shape (2,)
            Point that is to be interpolated.

        Returns
        -------
        out : int / float
            Interpolated value at `grid_pt`.
        """
        pts = w_pts.data
        dist = self._norm(pts[:, :2] - grid_pt)
        if np.any(dist == 0):
            mask = dist == 0
            return pts[mask][0, 2]

        wts = _set_weights(pts, dist)
        wts = np.square(wts)
        wts /= np.sum(wts)

        return wts @ pts[:, 2]


class ShepardInterpolator(Interpolator):
    """A full-featured inverse distance interpolator, based
    on the work of Shepard, "A two-dimensional interpolation
    function for irregularly-spaced data".


    Parameters
    ----------
    tol : positive float, optional
        The distance around every data point in which the data point is preferred to
        the interpolated data.

        Defaults to `numpy.finfo(float).eps`.

    slope : positive float, optional
        The initial slope used in Shepard`s algorithm.

        Defaults to `0.1`.

    norm : function or callable, optional
        Norm with which to calculate the distances, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled
        version of :math:`||.||_2`.


    Raises
    ------
    InterpolatorInitializationException
         If `tol` is non-positive.
    InterpolatorInitializationException
         If `slope` is non-positive.

    """

    def __init__(
        self,
        neighbourhood: Neighbourhood,
        tol=None,
        slope=0.1,
        norm: Callable = scaled_euclidean_norm,
    ):
        if tol is None:
            tol = np.finfo(float).eps
        super().__init__()
        if tol <= 0:
            raise InterpolatorInitializationException("`tol` is non-positive")
        if slope <= 0:
            raise InterpolatorInitializationException(
                "`slope` is non-positive"
            )

        self._norm = norm
        self._tol = tol
        self._slope = slope
        self._neighbourhood = neighbourhood

    def __repr__(self):
        return (
            f"ShepardInterpolator( tol={self._tol}, "
            f"slope_scal={self._slope}, norm={self._norm.__name__})"
        )

    def interpolate(self, w_pts, grid_pt):
        """Interpolates a given `grid_pt` according to the procedure described
        above.

        Parameters
        ----------
        w_pts : WeightedPoints
            Considered measured points.

        grid_pt : numpy.ndarray of shape (2,)
            Point that is to be interpolated.

        Returns
        -------
        out : int / float
            Interpolated value at `grid_pt`.
        """
        pts = w_pts.data
        dist = self._norm(pts[:, :2] - grid_pt)
        if np.any(dist == 0):
            mask = dist == 0
            return pts[mask][0, 2]

        wts = _set_weights(pts, dist)
        wts = _include_direction(pts, grid_pt, dist, wts)
        zdelta = _determine_slope(
            pts,
            grid_pt,
            dist,
            wts,
            self._neighbourhood,
            self._norm,
            self._slope,
        )

        # reduce computational errors
        mask = dist < self._tol
        n_eps = pts[mask]
        if n_eps.size:
            return np.sum(n_eps[:, 2]) / n_eps.shape[0]

        wts /= np.sum(wts)
        return wts @ (pts[:, 2] + zdelta)


def _set_weights(pts, dist):
    """"""
    wts = np.zeros(pts.shape[0])
    r = np.max(dist)

    for i, d in enumerate(dist):
        if 0 < d <= r / 3:
            wts[i] = 1 / d
        elif r / 3 < d <= r:
            wts[i] = 27 / (4 * r) * np.square(d / r - 1)

    return wts


def _include_direction(pts, grid_pt, dist, wts):
    """"""
    t = np.zeros(pts.shape[0])

    for i, pt in enumerate(pts):
        cosine = (grid_pt[0] - pt[0]) * (
            grid_pt[0] - np.delete(pts, i, axis=0)[:, 0]
        ) + (grid_pt[1] - pt[1]) * (
            grid_pt[1] - np.delete(pts, i, axis=0)[:, 0]
        )
        cosine /= dist[i] * np.delete(dist, i)
        t[i] = np.sum(np.delete(wts, i) * (1 - cosine))
        t[i] /= np.sum(wts)

    return np.square(wts) * (1 + t)


def _determine_slope(pts, grid_pt, dist, wts, nhood, norm, slope):
    """"""
    xderiv = np.zeros(pts.shape[0])
    yderiv = np.zeros(pts.shape[0])

    for i, pt in enumerate(pts):
        pts_i = np.delete(pts, i, axis=0)
        mask = nhood.is_contained_in(pts_i - pt)
        pts_i = pts_i[mask]
        dist_i = norm(pts_i - pt)
        wts_i = np.delete(wts, i)[mask]
        xderiv[i] = np.sum(
            wts_i
            * (pts_i[:, 2] - pt[2])
            * (pts_i[:, 0] - pt[0])
            / np.square(dist_i)
        )
        xderiv[i] /= np.sum(wts_i)
        yderiv[i] = np.sum(
            wts_i
            * (pts_i[:, 2] - pt[2])
            * (pts_i[:, 1] - pt[1])
            / np.square(dist_i)
        )
        yderiv[i] /= np.sum(wts_i)

    dim_of_dist = (
        slope
        * (np.max(pts[:, 2]) - np.min(pts[:, 2]))
        / np.sqrt(np.max(np.square(xderiv) + np.square(yderiv)))
    )

    return (
        xderiv * (grid_pt[0] - pts[:, 0]) + yderiv * (grid_pt[1] - pts[:, 1])
    ) * (dim_of_dist / (dim_of_dist + dist))
